SuperStrict

Module Pub.Mappy

ModuleInfo "Name: BlitzMax Mappy Module"
ModuleInfo "Description: BlitzMax port of the Mappy functions"
ModuleInfo "License: Public Domain"
ModuleInfo "Authors: gman (BlitzMax port), Robin Burrows (original blitz mappy functions)"

Import BRL.Pixmap
Import BRL.Max2D
Import BRL.FileSystem
Import BRL.StandardIO

Rem
 (C) Robin Burrows 2002 <robin@tilemap.co.uk>
 Please report bugs!
 For Mappy tilemaps: http://www.tilemap.co.uk
 Release 6 (29/12/2002) Full FMP1.0 support
 Lots of tiles may take a While To load with debug on, fast with debug off
EndRem

Public

Const MER_MAPTOONEW:Int = 3

' -------------------------------------------
' type for storing/restoring the current environment
' -------------------------------------------	
Type MAX2DENV
	Field vx:Int,vy:Int,vw:Int,vh:Int
	Field r:Int,g:Int,b:Int,blend:Int,alpha:Float
	
	Method New()
		EnvStore()
	EndMethod
	
	Method EnvStore()
		GetColor(r,g,b)
		blend=GetBlend()
		alpha=GetAlpha()
		GetViewport(vx,vy,vw,vh)
	EndMethod
	
	Method EnvRestore()
		SetColor(r,g,b)
		SetBlend(blend)
		SetAlpha(alpha)
		SetViewport(vx,vy,vw,vh)
	EndMethod
	
EndType

Type BLKSTR
	Field bg:Int, fg:Int, fg2:Int, fg3:Int
	Field user1:Int, user2:Int, user3:Int, user4:Int, user5:Int, user6:Int, user7:Int
	Field tl:Int, tr:Int, bl:Int, br:Int
	Field trigger:Int, unused1:Int, unused2:Int, unused3:Int
EndType

Type ANISTR
	Field antype:Int, andelay:Int, ancount:Int, anuser:Int
	Field ancur:Int, anstart:Int, anend:Int
EndType

Type MappyMap
	Field mapmaxlayers:Int=2
	Field maperror:Int
	Field mapislsb:Int, maptype:Int
	Field mapchunkdone:Int
	Field maptrans8:Int, maptransr:Int, maptransg:Int, maptransb:Int, maptransval:Int
	Field mapwidth:Int, mapheight:Int
	Field mapblockwidth:Int, mapblockheight:Int, mapdepth:Int
	Field mapblockstrsize:Int, mapnumblockstr:Int, mapnumblockgfx:Int
	Field mapblockgapx:Int, mapblockgapy:Int
	Field mapblockstaggerx:Int, mapblockstaggery:Int, mapclickmask_fld:Int
	Field mapcurlyr:Int, mapnumanims:Int

	Field mappt:Int[][][]
	Field aniseq:Int[]
	Field tilegfx:TImage[]
	Field cmappt:Int[]
	Field blks:BLKSTR[]
	Field anis:ANISTR[]
		
	Method MapFreeMem()
		Local n:Int
	
		For n = 0 To (mapnumblockstr-1)
			blks[n]=Null
		Next
		For n = 1 To (mapnumanims)
			anis[n]=Null
		Next
		For n = 0 To (mapnumblockgfx-1)
			tilegfx[n]=Null
		Next
		
		anis=anis:ANISTR[..0]
		blks=blks:BLKSTR[..0]
		tilegfx=tilegfx[..0]
		
	EndMethod

	Method MapGetBlock:BLKSTR(mx:Int,my:Int)
		If (mappt[mx][my][mapcurlyr] < 0)
			Return blks[aniseq[anis[-(mappt[mx][my][mapcurlyr])].ancur]]
		Else
			Return blks[mappt[mx][my][mapcurlyr]]
		End If
	EndMethod

	Method MapSetBlock(mx:Int,my:Int,mv:Int)
		mappt[mx][my][mapcurlyr]=mv
	EndMethod

	Method MapGetBlockInPixels:BLKSTR(mx:Int,my:Int)
		Local xp:Int, yp:Int
	
		If (mx < 0 Or my < 0 Or mx >= (mapblockwidth*mapwidth) Or my >= (mapblockheight*mapheight)) Then Return blks[0]
		xp = mx
		yp = my
		mx = MapGetXOffset(xp, yp)
		my = MapGetYOffset(xp, yp)
		
		If (mappt[mx][my][mapcurlyr] < 0)
			Return blks[aniseq[anis[-(mappt[mx][my][mapcurlyr])].ancur]]
		Else
			Return blks[mappt[mx][my][mapcurlyr]]
		End If
	EndMethod
	
	Method MapGetXOffset:Int(xpix:Int, ypix:Int)
		Local xb:Int
	
		If (mapblockstaggerx | mapblockstaggery)
			xpix = xpix+mapblockstaggerx
			ypix = ypix+mapblockstaggery
			xb = xpix/mapblockgapx
			If (mapclickmask_fld)
				xb = xb+MapClickMask(xpix, ypix, 0)
			End If
		Else
			xb = xpix/mapblockgapx
		End If
	
		If (xb < 0) Then xb = 0
		If (xb >= mapwidth) Then xb = mapwidth-1
		Return xb
	EndMethod
	
	Method MapGetYOffset:Int(xpix:Int, ypix:Int)
		Local yb:Int
	
		If (mapblockstaggerx | mapblockstaggery)
			xpix = xpix+mapblockstaggerx
			ypix = ypix+mapblockstaggery
			yb = ypix/mapblockgapy
			If (mapclickmask_fld)
				yb = yb*2
				yb = yb+MapClickMask(xpix, ypix, 1)
			End If
		Else
			yb = ypix/mapblockgapy
		End If
	
		If (yb < 0) Then yb = 0
		If (yb >= mapheight) Then yb = mapheight-1
		Return yb
	EndMethod

	Method MapClickMask:Int(x:Int, y:Int, xory:Int)
		x = x Mod mapblockgapx
		y = y Mod mapblockgapy
		If (x >= mapblockwidth And xory = 0) Then Return 0
		If (x >= mapblockwidth And xory = 1)
			If (y < mapblockstaggery)
				Return -1
			Else Return 1
			End If
		End If
		
		If (y >= mapblockheight And xory = 1) Then Return 1
		If (y >= mapblockheight And xory = 0)
			If (x < mapblockstaggerx)
				Return -1
			Else Return 0
			End If
		End If

		Local pixmap:TPixmap=LockImage(tilegfx[mapclickmask_fld])
		If (ReadPixel(pixmap,x, y) = maptransval)
			If (x < (mapblockwidth/2) And xory = 0) Then Return -1
			If (x >= (mapblockwidth/2) And xory = 0) Then Return 0
			If (y < (mapblockheight/2) And xory = 1) Then Return -1
			If (y >= (mapblockheight/2) And xory = 1) Then Return 1
		End If
		UnlockImage(tilegfx[mapclickmask_fld])
		Return 0
	EndMethod

	Method MapSetBlockInPixels(mx:Int, my:Int, mv:Int)
		If (mx < 0 Or my < 0 Or mx >= (mapblockwidth*mapwidth) Or my >= (mapblockheight*mapheight)) Then Return
	
		mappt[mx][my][mapcurlyr] = mv
	EndMethod

	Method MapCollision:Int(x:Int,y:Int)
		Local bRetVal:Int=False
		Local b:BLKSTR = MapGetBlock(x/mapblockwidth, y/mapblockheight) ' These must be valid map block coords
		If b.tl Then bRetVal=True
		Return bRetVal
	EndMethod
	
	' pixel-perfect collision checking against transparent pixels in a tile image
	Method MapPPCollision:Int(x:Int,y:Int)
		Local bRetVal:Int=False
		Local b:BLKSTR = MapGetBlock(x/mapblockwidth, y/mapblockheight) ' These must be valid map block coords
		If b.tl Then 
			Local pixmap:TPixmap=LockImage(tilegfx[b.fg])
			If (ReadPixel(pixmap,x Mod mapblockwidth, y Mod mapblockheight) <> maptransval) Then bRetVal=True
			UnlockImage(tilegfx[b.fg])
		End If
		Return bRetVal
	EndMethod

	Method MapLoad(mname:String,maxlayers:Int=2)
		Local fpt:TStream, maplen:Int, mapid1:Int, mapid2:Int, mapid3:Int, mapid4:Int, mapcklen:Int
		
		mapmaxlayers=maxlayers
		mapcurlyr = 0
		mapnumanims = 0
		mapblockstaggerx = 0
		mapblockstaggery = 0
		
		fpt = ReadFile(mname$)
		If MapReadByte(fpt) <> 70 maperror = 1
		If MapReadByte(fpt) <> 79 maperror = 1
		If MapReadByte(fpt) <> 82 maperror = 1
		If MapReadByte(fpt) <> 77 maperror = 1
		maplen = MapGetcksz (fpt)

		If MapReadByte(fpt) <> 70 maperror = 1
		If MapReadByte(fpt) <> 77 maperror = 1
		If MapReadByte(fpt) <> 65 maperror = 1
		If MapReadByte(fpt) <> 80 maperror = 1
		maplen = maplen - 4
	
		While (maplen)
			mapid1 = MapReadByte(fpt)
			mapid2 = MapReadByte(fpt)
			mapid3 = MapReadByte(fpt)
			mapid4 = MapReadByte(fpt)
			mapcklen = MapGetcksz(fpt)
		
			maplen = maplen - mapcklen
			maplen = maplen - 8
	
			mapchunkdone = 0
			If mapid1 = Asc("M") And mapid2 = Asc("P") And mapid3 = Asc("H") And mapid4 = Asc("D") Then MapDecodeMPHD(fpt, mapcklen)
			If mapid1 = Asc("C") And mapid2 = Asc("M") And mapid3 = Asc("A") And mapid4 = Asc("P") Then MapDecodeCMAP(fpt, mapcklen)
			If mapid1 = Asc("B") And mapid2 = Asc("K") And mapid3 = Asc("D") And mapid4 = Asc("T") Then MapDecodeBKDT(fpt, mapcklen)
			If mapid1 = Asc("A") And mapid2 = Asc("N") And mapid3 = Asc("D") And mapid4 = Asc("T") Then MapDecodeANDT(fpt, mapcklen)
			If mapid1 = Asc("B") And mapid2 = Asc("G") And mapid3 = Asc("F") And mapid4 = Asc("X") Then MapDecodeBGFX(fpt, mapcklen)
			If mapid1 = Asc("B") And mapid2 = Asc("O") And mapid3 = Asc("D") And mapid4 = Asc("Y") Then MapDecodeLYR(fpt, mapcklen, 0)
			If mapid1 = Asc("L") And mapid2 = Asc("Y") And mapid3 = Asc("R") And mapid4 = Asc("1") Then MapDecodeLYR(fpt, mapcklen, 1)
			If mapid1 = Asc("L") And mapid2 = Asc("Y") And mapid3 = Asc("R") And mapid4 = Asc("2") Then MapDecodeLYR(fpt, mapcklen, 2)
			If mapid1 = Asc("L") And mapid2 = Asc("Y") And mapid3 = Asc("R") And mapid4 = Asc("3") Then MapDecodeLYR(fpt, mapcklen, 3)
			If mapid1 = Asc("L") And mapid2 = Asc("Y") And mapid3 = Asc("R") And mapid4 = Asc("4") Then MapDecodeLYR(fpt, mapcklen, 4)
			If mapid1 = Asc("L") And mapid2 = Asc("Y") And mapid3 = Asc("R") And mapid4 = Asc("5") Then MapDecodeLYR(fpt, mapcklen, 5)
			If mapid1 = Asc("L") And mapid2 = Asc("Y") And mapid3 = Asc("R") And mapid4 = Asc("6") Then MapDecodeLYR(fpt, mapcklen, 6)
			If mapid1 = Asc("L") And mapid2 = Asc("Y") And mapid3 = Asc("R") And mapid4 = Asc("7") Then MapDecodeLYR(fpt, mapcklen, 7)

			If (mapchunkdone = 0)
				While (mapcklen)
					MapReadByte(fpt)
					mapcklen = mapcklen - 1
				Wend
			End If
	
		Wend
	
		CloseFile (fpt)
	EndMethod

	Method MapLoadMAR(mname:String)
		Local fpt:TStream, x:Int, y:Int, newf:Int
		
		fpt = ReadFile(mname$)
		newf = 0
		For y = 0 To (mapheight-1)
			For x = 0 To (mapwidth-1)
				mappt[x][y][mapcurlyr] = MapReadShort(fpt)
				If ((mappt[x][y][mapcurlyr] And 15) > 0) Then newf = 1
			Next
		Next
		CloseFile(fpt)
	
		If newf = 0
			For y = 0 To (mapheight-1)
				For x = 0 To (mapwidth-1)
					If (mappt[x][y][mapcurlyr] > 0)
						mappt[x][y][mapcurlyr] = mappt[x][y][mapcurlyr]/32
					Else
						mappt[x][y][mapcurlyr] = mappt[x][y][mapcurlyr]/16
					End If
				Next
			Next
		End If
	EndMethod

	Method MapGetcksz:Int(fcpt:TStream)
		Local l:Int
		l = ReadByte(fcpt)
		l = l * 256
		l = l + ReadByte(fcpt)
		l = l * 256
		l = l + ReadByte(fcpt)
		l = l * 256
		l = l + ReadByte(fcpt)
		Return l
	EndMethod
	
	Method MapInitAnims()
		Local n:Int
	
		If (mapnumanims < 1) Return
		n = 1
		While (anis[n].antype <> 255)
			If (anis[n].antype = 8) Then anis[n].antype = 5
			If (anis[n].antype = 7) Then anis[n].antype = 6
			If (anis[n].antype = 9) Then anis[n].antype = 3
			If (anis[n].antype = 2 Or anis[n].antype = 6)
				anis[n].ancur = anis[n].anstart
				If (anis[n].anstart <> anis[n].anend) Then anis[n].ancur = (anis[n].anend)-1
			Else
				anis[n].ancur = anis[n].anstart
			EndIf
	
			anis[n].ancount = anis[n].andelay
			n = n + 1
		Wend
	
	EndMethod
	
	Method MapUpdateAnims()
		Local n:Int
	
		If (mapnumanims < 1) Return
		n = 1
		While (anis[n].antype <> 255)
			If (anis[n].antype <> 0)
				anis[n].ancount = anis[n].ancount - 1
				If (anis[n].ancount < 0)
					anis[n].ancount = anis[n].andelay
					If (anis[n].anstart <> anis[n].anend)
						If (anis[n].antype = 1) ' LOOPF
							anis[n].ancur = anis[n].ancur + 1
							If (anis[n].ancur = anis[n].anend) Then anis[n].ancur = anis[n].anstart
						End If
						If (anis[n].antype = 2) ' LOOPR
							anis[n].ancur = anis[n].ancur - 1
							If (anis[n].ancur = ((anis[n].anend)-1)) Then anis[n].ancur = (anis[n].anend)-1
						End If
						If (anis[n].antype = 3) ' ONCE
							anis[n].ancur = anis[n].ancur + 1
							If (anis[n].ancur = anis[n].anend)
								anis[n].ancur = anis[n].anstart
								anis[n].antype = 9
							End If
						End If
						If (anis[n].antype = 4) ' ONCEH
							If (anis[n].ancur <> (anis[n].anend-1)) Then anis[n].ancur = anis[n].ancur + 1
						End If
						If (anis[n].antype = 5) ' PPFF
							anis[n].ancur = anis[n].ancur + 1
							If (anis[n].ancur = anis[n].anend) 
								anis[n].ancur = anis[n].ancur - 2
								anis[n].antype = 6
								If (anis[n].ancur < anis[n].anstart) Then anis[n].ancur = anis[n].ancur + 1
							End If
						Else
							If (anis[n].antype = 6) ' PPFR
								anis[n].ancur = anis[n].ancur - 1
								If (anis[n].ancur = (anis[n].anstart-1)) 
									anis[n].ancur = anis[n].ancur + 2
									anis[n].antype = 5
									If (anis[n].ancur > anis[n].anend) Then anis[n].ancur = anis[n].ancur - 1
								End If
							End If
						End If
						If (anis[n].antype = 7) ' PPRF
							anis[n].ancur = anis[n].ancur - 1
							If (anis[n].ancur = (anis[n].anstart-1)) 
								anis[n].ancur = anis[n].ancur + 2
								anis[n].antype = 6
								If (anis[n].ancur > anis[n].anend) Then anis[n].ancur = anis[n].ancur - 1
							End If
						Else
							If (anis[n].antype = 8) ' PPRR
								anis[n].ancur = anis[n].ancur + 1
								If (anis[n].ancur = anis[n].anend) 
									anis[n].ancur = anis[n].ancur - 2
									anis[n].antype = 5
									If (anis[n].ancur < anis[n].anstart) Then anis[n].ancur = anis[n].ancur + 1
								End If
							End If
						End If
					End If
				End If
			End If
			n = n + 1
		Wend
	
	EndMethod

	Method MapDecodeANDT(fcpt:TStream, chklen:Int) Private
		'DebugLog("ANDT")
		Local lastanim:Int, n:Int, strtpos:Int, andtpos:Int
	
		strtpos = StreamPos(fcpt)
		andtpos = chklen
	
		anis=anis:ANISTR[..1000+1]
	
		lastanim = 0
		n = 1
		While (lastanim = 0)
			andtpos = andtpos - 16
			SeekStream(fcpt, strtpos+andtpos)
			anis[n] = New ANISTR
			anis[n].antype = MapReadByte(fcpt)
			anis[n].andelay = MapReadByte(fcpt)
			anis[n].ancount = MapReadByte(fcpt)
			anis[n].anuser = MapReadByte(fcpt)
			anis[n].ancur = MapReadLong(fcpt)
			anis[n].anstart = MapReadLong(fcpt)
			anis[n].anend = MapReadLong(fcpt)
			If (maptype = 0)
				anis[n].ancur = (anis[n].ancur+chklen)/4
				anis[n].anstart = (anis[n].anstart+chklen)/4
				anis[n].anend = (anis[n].anend+chklen)/4
			End If
			If (anis[n].antype = 255) Then lastanim = 1
			n = n + 1
			mapnumanims = mapnumanims + 1
		Wend
		
		aniseq=aniseq[..andtpos/4+1]
		
		SeekStream(fcpt, strtpos)
		For n = 0 To ((andtpos-1)/4)
			aniseq[n] = MapReadLong(fcpt)
			If (maptype = 0) Then aniseq[n] = aniseq[n]/32
		Next
	
		SeekStream(fcpt, strtpos+chklen)
		mapchunkdone = 1
	EndMethod

	Method MapDecodeBKDT(fcpt:TStream, chklen:Int) Private
		'	Print "BKDT"
		'DebugLog("BKDT")
		Local n:Int, mbsize:Int, l:Int
	
		For n = 0 To (mapnumblockstr-1)
			blks[n] = New BLKSTR
			blks[n].bg = MapReadLong(fcpt)
			blks[n].fg = MapReadLong(fcpt)
			blks[n].fg2 = MapReadLong(fcpt)
			blks[n].fg3 = MapReadLong(fcpt)
			If (maptype = 0)
				If (mapdepth = 15) 
					mbsize = (mapblockwidth*mapblockheight*2)
				Else
					mbsize = (mapblockwidth*mapblockheight*(mapdepth/8))
				End If
				blks[n].bg = blks[n].bg/mbsize
				blks[n].fg = blks[n].fg/mbsize
				blks[n].fg2 = blks[n].fg2/mbsize
				blks[n].fg3 = blks[n].fg3/mbsize
			End If
			blks[n].user1 = MapReadLong(fcpt)
			blks[n].user2 = MapReadLong(fcpt)
			blks[n].user3 = MapReadShort(fcpt)
			blks[n].user4 = MapReadShort(fcpt)
			blks[n].user5 = MapReadByte(fcpt)
			blks[n].user6 = MapReadByte(fcpt)
			blks[n].user7 = MapReadByte(fcpt)
			l = MapReadByte(fcpt)
			
			blks[n].tl = 0
			If (l & 1) Then blks[n].tl = 1
			blks[n].tr = 0
			If (l & 2) Then blks[n].tr = 1
			blks[n].bl = 0
			If (l & 4) Then blks[n].bl = 1
			blks[n].br = 0
			If (l & 8) Then blks[n].br = 1
			blks[n].trigger = 0
			If (l & 16) Then blks[n].trigger = 1
			blks[n].unused1 = 0
			If (l & 32) Then blks[n].unused1 = 1
			blks[n].unused2 = 0
			If (l & 64) Then blks[n].unused2 = 1
			blks[n].unused3 = 0
			If (l & 128) Then blks[n].unused3 = 1
			chklen = chklen-32
		Next
	
		While (chklen)
			MapReadByte(fcpt)
			chklen = chklen - 1
		Wend
		mapchunkdone = 1
	EndMethod

	Method MapDecodeMPHD:Int(fcpt:TStream, chklen:Int) Private
		'	Print "MPHD"
		'DebugLog("MPHD")
		If (MapReadByte(fcpt) > 1)
			maperror = MER_MAPTOONEW
			Return -1
		End If
		MapReadByte(fcpt)
		If (MapReadByte(fcpt) = 1)
			mapislsb = 1
		Else
			mapislsb = 0
		End If
	
		maptype = MapReadByte(fcpt)
		If (maptype > 3)
			maperror = MER_MAPTOONEW
			Return -1
		End If
		
		chklen = chklen-4
		mapwidth = MapReadShort(fcpt)
		mapheight = MapReadShort(fcpt)
		MapReadShort(fcpt)
		MapReadShort(fcpt)
		mapblockwidth = MapReadShort(fcpt)
		mapblockheight = MapReadShort(fcpt)
		mapdepth = MapReadShort(fcpt)
		mapblockstrsize = MapReadShort(fcpt)
		mapnumblockstr = MapReadShort(fcpt)
		mapnumblockgfx = MapReadShort(fcpt)
				
		ResizeMAPPT(mapwidth+1,mapheight+1,mapmaxlayers+1)
		
		tilegfx=tilegfx[..mapnumblockgfx+1]
		blks=blks:BLKSTR[..mapnumblockstr+1]
	
		chklen = chklen-20
		If (chklen > 0)
			maptrans8 = MapReadByte(fcpt)
			maptransr = MapReadByte(fcpt)
			maptransg = MapReadByte(fcpt)
			maptransb = MapReadByte(fcpt)
			chklen = chklen-4
			If (chklen > 0)
				mapblockgapx = MapReadShort(fcpt)
				mapblockgapy = MapReadShort(fcpt)
				mapblockstaggerx = MapReadShort(fcpt)
				mapblockstaggery = MapReadShort(fcpt)
				chklen = chklen-8
				If (chklen > 0)
					mapclickmask_fld = MapReadShort(fcpt)
					chklen = chklen-2
				Else
					mapclickmask_fld = 0
				End If
			Else
				mapblockgapx = mapblockwidth
				mapblockgapy = mapblockheight
				mapblockstaggerx = 0
				mapblockstaggery = 0
			End If
		Else 
			maptrans8 = 0
			maptransr = 255
			maptransg = 0
			maptransb = 255
		End If
		
		While (chklen)
			MapReadByte(fcpt)
			chklen = chklen - 1
		Wend
		mapchunkdone = 1
	EndMethod

	Method MapDecodeLYR(fcpt:TStream, chklen:Int, lyrnum:Int) Private
	'	Print "LYR"+(lyrnum)
		'DebugLog("LYR"+lyrnum)
		Local n:Int, x:Int, y:Int, rlecount:Int, rleval:Int
				
		If (maptype = 0)
			For y = 0 To (mapheight-1)
				For x = 0 To (mapwidth-1)
					mappt[x][y][lyrnum] = MapReadShort(fcpt)
					If (mappt[x][y][lyrnum] < 0)
						mappt[x][y][lyrnum] = mappt[x][y][lyrnum]/16
					Else
						mappt[x][y][lyrnum] = mappt[x][y][lyrnum]/mapblockstrsize
					EndIf
					chklen = chklen-2
				Next
			Next
		EndIf
		If (maptype = 1)
			For y = 0 To (mapheight-1)
				For x = 0 To (mapwidth-1)
					mappt[x][y][lyrnum] = MapReadShort(fcpt)
					chklen = chklen-2
				Next
			Next
		End If
		If (maptype = 2)
			For y = 0 To (mapheight-1)
				x = 0 ; While (x < mapwidth)
					rlecount = MapReadShort(fcpt)
					chklen = chklen-2
					If (rlecount > 0)
						While (rlecount > 0)
							mappt[x][y][lyrnum] = MapReadShort(fcpt)
							chklen = chklen-2
							x = x + 1 ; rlecount = rlecount - 1
						Wend
					Else
						rleval = MapReadShort(fcpt)
						chklen = chklen-2
						While (rlecount < 0)
							mappt[x][y][lyrnum] = rleval
							x = x + 1 ; rlecount = rlecount + 1
						Wend
					End If
				Wend
			Next
		End If
		If (maptype > 2)
			Print "Sorry, unsupported use FMP0.5 or FMP1.0 for now"
			DebugStop
		End If
	
		While (chklen)
			MapReadByte(fcpt)
			chklen = chklen - 1
		Wend
		mapchunkdone = 1
	EndMethod

	Method MapDecodeCMAP(fcpt:TStream, chklen:Int) Private
	'	Print "CMAP"
		'DebugLog("CMAP")
		Local n:Int
	
		cmappt=cmappt[..chklen+1]
		For n = 0 To (chklen-1)
			cmappt[n] = MapReadByte(fcpt)
		Next
		mapchunkdone = 1
	EndMethod

	Method MapReadLong:Int(fcpt:TStream) Private
		Local l:Int
		If (mapislsb = 1)
			l = ReadByte(fcpt)
			l = l + (ReadByte(fcpt)*256)
			l = l + (ReadByte(fcpt)*65536)
			l = l + (ReadByte(fcpt)*16777216)
		Else
			l = ReadByte(fcpt)
			l = l * 256
			l = l + ReadByte(fcpt)
			l = l * 256
			l = l + ReadByte(fcpt)
			l = l * 256
			l = l + ReadByte(fcpt)
		EndIf
		Return l
	EndMethod
	
	Method MapReadShort:Int(fcpt:TStream) Private
		Local l:Int
		If (mapislsb = 1)
			l = ReadByte(fcpt)
			l = l + (ReadByte(fcpt)*256)
		Else
			l = ReadByte(fcpt)
			l = l * 256
			l = l + ReadByte(fcpt)
		EndIf
		If (l > 32767) Then l = l -65536
		Return l
	EndMethod
	
	Method MapReadByte:Int(fcpt:TStream) Private
		Return ReadByte(fcpt)
	EndMethod

	Method ResizeMAPPT(width:Int, height:Int, depth:Int) Private
		Local i:Int,j:Int
				
		' Resize the main array (which is used here as the width-index)
		mappt = mappt[..width]
		
		' Resize all subarrays (each index of the main array points to a different subarray)
		For i = 0 Until width ' "Until" is used, because you have index "0" to index "width - 1" (0-based arrays)
			mappt[i] = mappt[i][..height]
		Next
	
		For i = 0 Until width ' "Until" is used, because you have index "0" to index "width - 1" (0-based arrays)
			For j=0 Until height
				mappt[i][j] = mappt[i][j][..depth]
			Next
		Next
	EndMethod

	Method MapDecodeBGFX(fcpt:TStream, chklen:Int) Private
	   '	Print "BGFX"
		'DebugLog("BGFX")
	   Local n:Int, x:Int, y:Int, p:Int, p2:Int, pred:Int, pgreen:Int, pblue:Int
		Local pixmap:TPixmap
	
		For n = 0 To (mapnumblockgfx-1)
		
			pixmap = CreatePixmap(mapblockwidth, mapblockheight, PF_RGBA8888)
			
			If (n = 0)
				If (mapdepth = 8)
					p = maptrans8
					WritePixel(pixmap, 0, 0, (cmappt[p*3] Shl 16) | (cmappt[(p*3)+1] Shl 8) | (cmappt[(p*3)+2]))
				Else
					WritePixel(pixmap, 0, 0, (maptransr Shl 16) | (maptransg Shl 8) | maptransb)
				End If
				maptransval = ReadPixel(pixmap,0, 0)				
				tilegfx[n]=LoadImage(pixmap)				
			End If
			If (mapdepth = 8)
				For y = 0 To (mapblockheight-1)
					For x = 0 To (mapblockwidth-1)
						p = MapReadByte(fcpt)
						chklen = chklen - 1
						
						If (cmappt[p*3]=maptransr And cmappt[(p*3)+1]=maptransg And cmappt[(p*3)+2]=maptransb) Or ..
							((cmappt[p*3] Shl 16) | (cmappt[(p*3)+1] Shl 8) | cmappt[(p*3)+2])=maptransval Or ..
							((cmappt[p*3] Shl 16) | (cmappt[(p*3)+1] Shl 8) | cmappt[(p*3)+2])=maptrans8
														
							WritePixel(pixmap, x, y, GetARGB(0,cmappt[p*3],cmappt[(p*3)+1],cmappt[(p*3)+2]))
						Else
							WritePixel(pixmap, x, y, GetARGB(255,cmappt[p*3],cmappt[(p*3)+1],cmappt[(p*3)+2]))
						EndIf						
					Next
				Next
				tilegfx[n]=LoadImage(pixmap)
			End If
			If (mapdepth = 15)
				For y = 0 To (mapblockheight-1)
					For x = 0 To (mapblockwidth-1)
						p = MapReadByte(fcpt)
						p2 = MapReadByte(fcpt)
						pred = (p Shr 2) Shl 3
						pred = pred | (pred Shr 5)
						pgreen = (p & 3) Shl 6
						pgreen = pgreen | ((p2 Shr 5) Shl 3)
						pgreen = pgreen | (pgreen Shr 5)
						pblue = (p2 Shl 3) & 255
						pblue = pblue | (pblue Shr 5)
						chklen = chklen - 2
						If ((pred Shl 16) | (pgreen Shl 8) | pblue)=maptransval							
							WritePixel(pixmap,x, y, GetARGB(0,pred,pgreen,pblue))
						Else
							WritePixel(pixmap,x, y, GetARGB(255,pred,pgreen,pblue))
						EndIf
					Next
				Next
				tilegfx[n]=LoadImage(pixmap)
			End If
			If (mapdepth = 16)
				For y = 0 To (mapblockheight-1)
					For x = 0 To (mapblockwidth-1)
						p = MapReadByte(fcpt)
						p2 = MapReadByte(fcpt)
						pred = (p Shr 3) Shl 3
						pred = pred | (pred Shr 5)
						pgreen = (p & 7) Shl 5
						pgreen = pgreen | ((p2 Shr 5) Shl 2)
						pgreen = pgreen | (pgreen Shr 6)
						pblue = (p2 Shl 3) & 255
						pblue = pblue | (pblue Shr 5)
						chklen = chklen - 2
						If ((pred Shl 16) | (pgreen Shl 8) | pblue)=maptransval						
							WritePixel(pixmap, x, y, GetARGB(0,pred,pgreen,pblue))
						Else
							WritePixel(pixmap, x, y, GetARGB(255,pred,pgreen,pblue))
						EndIf
					Next
				Next
				tilegfx[n]=LoadImage(pixmap)
			End If
			If (mapdepth = 24)
				For y = 0 To (mapblockheight-1)
					For x = 0 To (mapblockwidth-1)
						pred = MapReadByte(fcpt)
						pgreen = MapReadByte(fcpt)
						pblue = MapReadByte(fcpt)
						chklen = chklen - 3
						If ((pred Shl 16) | (pgreen Shl 8) | pblue)=maptransval						
							WritePixel(pixmap, x, y, GetARGB(0,pred,pgreen,pblue))
						Else
							WritePixel(pixmap, x, y, GetARGB(255,pred,pgreen,pblue))
						EndIf
					Next
				Next
				tilegfx[n]=LoadImage(pixmap)
			End If
			If (mapdepth = 32)
				For y = 0 To (mapblockheight-1)
					For x = 0 To (mapblockwidth-1)
						MapReadByte(fcpt)
						pred = MapReadByte(fcpt)
						pgreen = MapReadByte(fcpt)
						pblue = MapReadByte(fcpt)
						chklen = chklen - 4
						If ((pred Shl 16) | (pgreen Shl 8) | pblue)=maptransval
							WritePixel(pixmap, x, y, GetARGB(0,pred,pgreen,pblue))
						Else
							WritePixel(pixmap, x, y, GetARGB(255,pred,pgreen,pblue))
						EndIf
					Next
				Next
				tilegfx[n]=LoadImage(pixmap)
			End If
		Next
		
		While (chklen)
			MapReadByte(fcpt)
			chklen = chklen - 1
		Wend
		mapchunkdone = 1
	EndMethod

	' -------------------------------------------------------------------
	' drawing methods
	' -------------------------------------------------------------------
				
	Method MapDrawRow(mx:Int, my:Int, sx:Int, sy:Int, sw:Int, sh:Int, rw:Int)
		Local tileg:Int, sy2:Int, attblk:Int, attloop:Int
	
		If (((my/mapblockgapy)+rw) >= mapheight) Then Return 
	
		' store the current 2D drawing environment
		Local env:MAX2DENV=New MAX2DENV
		SetDrawingEnv()
		
		SetViewport(sx, sy, sw, sh)
		sw = sw + sx
		sh = sh + sy
		
		sx = sx -(mx Mod mapblockgapx)
		sy = sy -(my Mod mapblockgapy)
		sy = sy + ((rw/2)*mapblockgapy)
		If (Not (rw & 1))
			sx = sx - mapblockstaggerx
			sy = sy - mapblockstaggery
		End If
		mx = mx / mapblockgapx
		my = (my / mapblockgapy) * 2
		my = my + rw
	
		sy2 = sy
		
		While (sx < sw)
			sy = sy2
			attloop = 0
			Repeat
				If (mappt[mx][my][mapcurlyr] < 0)
					attblk = aniseq[anis[-(mappt[mx][my][mapcurlyr])].ancur]+attloop
				Else
					attblk = mappt[mx][my][mapcurlyr]+attloop
				End If
				If (attloop > 0)
					tileg = blks[attblk].bg
					If (tileg > 0)
						If ((blks[attblk].unused2 + blks[attblk].unused3) = 0) Then MapDrawImageSimple(tileg, sx, sy)
						If (blks[attblk].unused2 & (Not (blks[attblk].unused3))) Then MapDrawImageRect(tileg, sx, sy, 0, 0, mapblockwidth/2, mapblockheight)
						If (blks[attblk].unused3 & (Not (blks[attblk].unused2))) Then MapDrawImageRect(tileg, sx+(mapblockwidth/2), sy, mapblockwidth/2, 0, mapblockwidth/2, mapblockheight)
					End If
					sy = sy - mapblockheight
				End If
				tileg = blks[attblk].fg
				If (tileg > 0)
					If ((blks[attblk].unused2 + blks[attblk].unused3) = 0) Then MapDrawImageSimple(tileg, sx, sy)
					If (blks[attblk].unused2 & (Not (blks[attblk].unused3))) Then MapDrawImageRect(tileg, sx, sy, 0, 0, mapblockwidth/2, mapblockheight)
					If (blks[attblk].unused3 & (Not (blks[attblk].unused2))) Then MapDrawImageRect(tileg, sx+(mapblockwidth/2), sy, mapblockwidth/2, 0, mapblockwidth/2, mapblockheight)
				End If
				sy = sy - mapblockheight
				tileg = blks[attblk].fg2
				If (tileg > 0)
					If ((blks[attblk].unused2 + blks[attblk].unused3) = 0) Then MapDrawImageSimple(tileg, sx, sy)
					If (blks[attblk].unused2 & (Not (blks[attblk].unused3))) Then MapDrawImageRect(tileg, sx, sy, 0, 0, mapblockwidth/2, mapblockheight)
					If (blks[attblk].unused3 & (Not (blks[attblk].unused2))) Then MapDrawImageRect(tileg, sx+(mapblockwidth/2), sy, mapblockwidth/2, 0, mapblockwidth/2, mapblockheight)
				End If
				sy = sy - mapblockheight
				tileg = blks[attblk].fg3
				If (tileg > 0)
					If ((blks[attblk].unused2 + blks[attblk].unused3) = 0) Then MapDrawImageSimple(tileg, sx, sy)
					If (blks[attblk].unused2 & (Not (blks[attblk].unused3))) Then MapDrawImageRect(tileg, sx, sy, 0, 0, mapblockwidth/2, mapblockheight)
					If (blks[attblk].unused3 & (Not (blks[attblk].unused2))) Then MapDrawImageRect(tileg, sx+(mapblockwidth/2), sy, mapblockwidth/2, 0, mapblockwidth/2, mapblockheight)
				End If
				sy = sy - mapblockheight
				attloop = attloop + 1
			Until (blks[attblk].unused1 = 0)
			mx = mx + 1
			sx = sx + mapblockgapx
		Wend
		
		env.EnvRestore()
	EndMethod
	
	Method MapDrawFG (mx:Int, my:Int, sx:Int, sy:Int, sw:Int, sh:Int, blyr:Int)
		Local mx2:Int, sx2:Int, tileg:Int
	
		If (mapblockstaggerx | mapblockstaggery)
			MapDrawStagger(mx, my, sx, sy, sw, sh, blyr+1)
			Return
		EndIf
	
		' store the current 2D drawing environment
		Local env:MAX2DENV=New MAX2DENV
		SetDrawingEnv()
	
		SetViewport(sx, sy, sw, sh)
		sw = sw + sx
		sh = sh + sy
		
		sx = sx -(mx Mod mapblockwidth)
		sy = sy -(my Mod mapblockheight)
		mx = mx / mapblockwidth
		my = my / mapblockheight
		mx2 = mx
		sx2 = sx
	
		While (sy < sh)
			mx = mx2
			sx = sx2
	
			While (sx < sw)
				If (mappt[mx][my][mapcurlyr] < 0)
					If (blyr = 0) tileg = blks[aniseq[anis[-(mappt[mx][my][mapcurlyr])].ancur]].fg
					If (blyr = 1) tileg = blks[aniseq[anis[-(mappt[mx][my][mapcurlyr])].ancur]].fg2
					If (blyr = 2) tileg = blks[aniseq[anis[-(mappt[mx][my][mapcurlyr])].ancur]].fg3
				Else
					If (blyr = 0) tileg = blks[mappt[mx][my][mapcurlyr]].fg
					If (blyr = 1) tileg = blks[mappt[mx][my][mapcurlyr]].fg2
					If (blyr = 2) tileg = blks[mappt[mx][my][mapcurlyr]].fg3
				End If
				If (tileg > 0) Then MapDrawImageSimple(tileg, sx, sy)
				mx = mx + 1
				sx = sx + mapblockwidth
			Wend
			my = my + 1
			sy = sy + mapblockheight
		Wend

		' restore the environment	
		env.EnvRestore()
	EndMethod

	Method MapDrawBGT(mx:Int, my:Int, sx:Int, sy:Int, sw:Int, sh:Int)
		Local mx2:Int, sx2:Int, tileg:Int
	
		If (mapblockstaggerx | mapblockstaggery)
			MapDrawStagger(mx, my, sx, sy, sw, sh, 0)
			Return
		EndIf

		' store the current 2D drawing environment
		Local env:MAX2DENV=New MAX2DENV
		SetDrawingEnv()
				
		SetViewport(sx, sy, sw, sh)
		sw = sw + sx
		sh = sh + sy
		
		sx = sx -(mx Mod mapblockwidth)
		sy = sy -(my Mod mapblockheight)
		mx = mx / mapblockwidth
		my = my / mapblockheight
		mx2 = mx
		sx2 = sx
	
		While (sy < sh)
			mx = mx2
			sx = sx2
	
			While (sx < sw)
				If (mappt[mx][my][mapcurlyr] < 0)
					tileg = blks[aniseq[anis[-(mappt[mx][my][mapcurlyr])].ancur]].bg
				Else
					tileg = blks[mappt[mx][my][mapcurlyr]].bg
				End If
				If (tileg > 0) Then MapDrawImageSimple(tileg, sx, sy)
				mx = mx + 1
				sx = sx + mapblockwidth
			Wend
			my = my + 1
			sy = sy + mapblockheight
		Wend

		' restore the environment
		env.EnvRestore()	
	EndMethod

	Method MapDrawBG(mx:Int, my:Int, sx:Int, sy:Int, sw:Int, sh:Int)
		Local mx2:Int, sx2:Int
	
		If (mapblockstaggerx | mapblockstaggery)
			MapDrawStagger(mx, my, sx, sy, sw, sh, 0)
			Return
		End If

		' store the current 2D drawing environment
		Local env:MAX2DENV=New MAX2DENV
		SetDrawingEnv()
				
		SetViewport(sx, sy, sw, sh)
		sw = sw + sx
		sh = sh + sy
		
		sx = sx -(mx Mod mapblockwidth)
		sy = sy -(my Mod mapblockheight)
		mx = mx / mapblockwidth
		my = my / mapblockheight
		mx2 = mx
		sx2 = sx
	
		While (sy < sh)
			mx = mx2
			sx = sx2
	
			While (sx < sw)
				If (mappt[mx][my][mapcurlyr] < 0)
					MapDrawImageSimple(blks[aniseq[anis[-(mappt[mx][my][mapcurlyr])].ancur]].bg, sx, sy, 0, True)
				Else
					MapDrawImageSimple(blks[mappt[mx][my][mapcurlyr]].bg, sx, sy, 0, True)
				End If
				mx = mx + 1
				sx = sx + mapblockwidth
			Wend
			my = my + 1
			sy = sy + mapblockheight
		Wend	
		
		' restore the environment
		env.EnvRestore()
	EndMethod

	Method MapDrawStagger(mx:Int, my:Int, sx:Int, sy:Int, sw:Int, sh:Int, bl:Int)
		Local mx2:Int, sx2:Int

		' store the current 2D drawing environment
		Local env:MAX2DENV=New MAX2DENV
		SetDrawingEnv()
		
		SetViewport(sx, sy, sw, sh)
		sw = sw + sx
		sh = sh + sy
		
		sx = sx -(mx Mod mapblockgapx)
		sx = sx - mapblockstaggerx
		sy = sy -(my Mod mapblockgapy)
		sy = sy - mapblockstaggery
		mx = mx / mapblockgapx
		my = (my / mapblockgapy) * 2
		mx2 = mx
		sx2 = sx
	
		While (sy < sh)
			mx = mx2
			sx = sx2
	
			While (sx < sw)
				If (mappt[mx][my][mapcurlyr] < 0)
					If (bl = 0) MapDrawImageSimple(blks[aniseq[anis[-(mappt[mx][my][mapcurlyr])].ancur]].bg, sx, sy)
					If (bl = 1) MapDrawImageSimple(blks[aniseq[anis[-(mappt[mx][my][mapcurlyr])].ancur]].fg, sx, sy)
					If (bl = 2) MapDrawImageSimple(blks[aniseq[anis[-(mappt[mx][my][mapcurlyr])].ancur]].fg2, sx, sy)
					If (bl = 3) MapDrawImageSimple(blks[aniseq[anis[-(mappt[mx][my][mapcurlyr])].ancur]].fg3, sx, sy)
				Else
					If (bl = 0) MapDrawImageSimple(blks[mappt[mx][my][mapcurlyr]].bg, sx, sy)
					If (bl = 1) MapDrawImageSimple(blks[mappt[mx][my][mapcurlyr]].fg, sx, sy)
					If (bl = 2) MapDrawImageSimple(blks[mappt[mx][my][mapcurlyr]].fg2, sx, sy)
					If (bl = 3) MapDrawImageSimple(blks[mappt[mx][my][mapcurlyr]].fg3, sx, sy)
				End If
				mx = mx + 1
				sx = sx + mapblockgapx
			Wend
			my = my + 1
			mx = mx2
			sx = sx2+mapblockstaggerx
			sy = sy+mapblockstaggery
			While (sx < sw)
				If (mappt[mx][my][mapcurlyr] < 0)
					If (bl = 0) MapDrawImageSimple(blks[aniseq[anis[-(mappt[mx][my][mapcurlyr])].ancur]].bg, sx, sy)
					If (bl = 1) MapDrawImageSimple(blks[aniseq[anis[-(mappt[mx][my][mapcurlyr])].ancur]].fg, sx, sy)
					If (bl = 2) MapDrawImageSimple(blks[aniseq[anis[-(mappt[mx][my][mapcurlyr])].ancur]].fg2, sx, sy)
					If (bl = 3) MapDrawImageSimple(blks[aniseq[anis[-(mappt[mx][my][mapcurlyr])].ancur]].fg3, sx, sy)
				Else
					If (bl = 0) MapDrawImageSimple(blks[mappt[mx][my][mapcurlyr]].bg, sx, sy)
					If (bl = 1) MapDrawImageSimple(blks[mappt[mx][my][mapcurlyr]].fg, sx, sy)
					If (bl = 2) MapDrawImageSimple(blks[mappt[mx][my][mapcurlyr]].fg2, sx, sy)
					If (bl = 3) MapDrawImageSimple(blks[mappt[mx][my][mapcurlyr]].fg3, sx, sy)
				End If
				mx = mx + 1
				sx = sx + mapblockgapx
			Wend
			my = my + 1
			sy = sy + (mapblockgapy-mapblockstaggery)
		Wend

		' restore the environment
		env.EnvRestore()	
	EndMethod

	' draws an image without worrying about the current 2D environment
	Method MapDrawImageSimple(idx:Int,x:Float,y:Float,frame:Int=0,bDrawBlock:Int=False)
		If Not bDrawBlock Then DrawImage(tilegfx[idx],x,y) Else DrawBlock(tilegfx[idx],x,y)
	EndMethod

	' draws an image.  stores/restores the current 2D environment																																				
	Method MapDrawImage(idx:Int,x:Float,y:Float,frame:Int=0,bDrawBlock:Int=False)
		' store the current 2D drawing environment
		Local env:MAX2DENV=New MAX2DENV	
		
		SetDrawingEnv()
		
		MapDrawImageSimple(idx,x,y,frame,bDrawBlock)
		
		env.EnvRestore()
	EndMethod

	' draws a portion of the image
	Method MapDrawImageRect(idx:Int,x:Float,y:Float,rx:Float,ry:Float,rw:Float,rh:Float,frame:Int=0)
		' store the current 2D drawing environment
		Local env:MAX2DENV=New MAX2DENV	

		SetDrawingEnv()
						
		SetViewport(x,y,rw,rh)
		
		DrawImage(tilegfx[idx],x-rx,x-ry,frame)

		env.EnvRestore()						
	EndMethod
	
	' sets the 2D drawing environment for mappy
	Method SetDrawingEnv()
		SetAlpha(1)
		SetColor(255,255,255)
		SetBlend(ALPHABLEND)	
	EndMethod
EndType

Private

Function GetARGB:Int(a:Int,r:Int,g:Int,b:Int)
	Return ((a & $ff) Shl 24) | ((r & $ff) Shl 16) | ((g & $ff) Shl 8) | (b & $ff)
EndFunction

Function DrawBlock(Image:TImage, x#, y#, frame:Int=0)
	Local oldBlend:Int = GetBlend()
	SetBlend(SOLIDBLEND)
	DrawImage(Image, X, Y, frame)
	SetBlend(oldBlend)
End Function
